/* 
 * wrapInit.c --
 *
 *	This file contains the initialization code, and the top-level
 *	code for the "::wrap::source", "::wrap::open", and "::wrap::info" commands.
 *
 * Copyright (c) 1999      Jan Nijtmans
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <string.h>
#include <stdlib.h>

#include "wrap.h"

/*
 * Declarations for functions defined in this file.
 */

#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN
#   if defined(_MSC_VER)
#	define EXPORT(a,b) __declspec(dllexport) a b
#   else
#	if defined(__BORLANDC__)
#	    define EXPORT(a,b) a _export b
#	else
#	    define EXPORT(a,b) a b
#	endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif

EXTERN EXPORT(int,Wrap_Init) _ANSI_ARGS_((Tcl_Interp *interp));

static int wrapSource _ANSI_ARGS_((ClientData clientData,
	Tcl_Interp *interp, int objc, struct Tcl_Obj * CONST objv[]));
static int wrapOpen _ANSI_ARGS_((ClientData clientData,
	Tcl_Interp *interp, int objc, struct Tcl_Obj * CONST objv[]));
static int wrapInfo _ANSI_ARGS_((ClientData clientData,
	Tcl_Interp *interp, int objc, struct Tcl_Obj * CONST objv[]));
static void wrapDeleteProc _ANSI_ARGS_((ClientData clientData));

/*
 *--------------------------------------------------------------
 *
 * Wrap_Init --
 *	Create Wrap commands.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static Tcl_Obj *resource = NULL;

static char initScript[] = "if {[info proc wrapInit]==\"\"} {\n\
  proc wrapInit {} {\n\
    global ::wrap::library ::wrap::version ::wrap::patchLevel\n\
    rename wrapInit {}\n\
    if {![info exists ::wrap::library]} {\n\
	set ::wrap::library {}\n\
    }\n\
    tcl_findLibrary wrap ${::wrap::version} ${::wrap::patchLevel} wrap.tcl WRAP_LIBRARY ::wrap::library\n\
  }\n\
}\n\
if {[info exists auto_path]} {\n\
  wrapInit\n\
}";


EXPORT(int,Wrap_Init)(interp)
    Tcl_Interp *interp;
{
    if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
	return TCL_ERROR;
    }
    if (!resource) {
	resource = Tcl_NewStringObj((char *) Tcl_GetNameOfExecutable(), -1);	
    }

    Tcl_CreateObjCommand(interp,"::wrap::source", wrapSource,
	    (ClientData) NULL, wrapDeleteProc);
    Tcl_IncrRefCount(resource);

    Tcl_CreateObjCommand(interp,"::wrap::open", wrapOpen,
	    (ClientData) NULL, wrapDeleteProc);
    Tcl_IncrRefCount(resource);

    Tcl_CreateObjCommand(interp,"::wrap::info", wrapInfo, (ClientData) NULL, NULL);

    Tcl_SetVar(interp, "::wrap::patchLevel", WRAP_PATCH_LEVEL, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "::wrap::version", WRAP_VERSION, TCL_GLOBAL_ONLY);

    if (Tcl_Eval(interp, initScript) != TCL_OK) {
	return TCL_ERROR;
    }
    return Tcl_PkgProvide(interp,"Wrap",WRAP_VERSION);
}

/*
 *-------------------------------------------------------------------------
 *
 * wrapSource --
 *
 *	This procedure is invoked to process the "::wrap::source" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 * *-------------------------------------------------------------------------
 */

static int wrapSource(clientData, interp, objc, objv)
    ClientData clientData;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST objv[];
{
    Tcl_DString dstring;
    char *string;
    int num, length;
    Tcl_Obj *obj;
    Tcl_Channel chan;
    Tcl_Obj *rsrc = resource;
    int result = TCL_OK;
    CONST char *entry;

    if (objc < 2 || objc > 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "fileName ?zipFileName?");
	return TCL_ERROR;
    }
    if (objc > 2) {
	rsrc = objv[2];
    }
    entry = Tcl_GetString(objv[1]);
    chan = Wrap_Open(interp, rsrc, entry);
    if (!chan)
	return TCL_ERROR;
    obj = Tcl_NewObj();
    Tcl_IncrRefCount(obj);
    if (Tcl_ReadChars(chan, obj, -1, 0) < 0) {
        Tcl_Close(interp, chan);
	Tcl_AppendResult(interp, "couldn't read entry \"", entry,
		"\": ", Tcl_PosixError(interp), (char *) NULL);
	goto done;
    }
    if (Tcl_Close(interp, chan) != TCL_OK) {
        goto done;
    }

    string = Tcl_GetStringFromObj(obj, &length);
    result = Tcl_EvalEx(interp, string, length, TCL_EVAL_DIRECT);

done:
    if (result == TCL_RETURN) {
	result = TCL_OK;
    } else if (result == TCL_ERROR) {
	char msg[400];

	/*
	 * Record information telling where the error occurred.
	 */

	sprintf(msg, "\n    (file: \"%.150s\" entry: \"%.150s\" line %d)",
		Tcl_GetString(rsrc), entry, interp->errorLine);
	Tcl_AddErrorInfo(interp, msg);
    }

    Tcl_DecrRefCount(obj);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * wrapOpen --
 *
 *	This procedure is invoked to process the "::wrap::open" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */

static int wrapOpen(clientData, interp, objc, objv)
    ClientData clientData;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST objv[];
{
    Tcl_Channel chan;
    char *fileName;
    Tcl_Obj *rsrc = resource;
    char buffer[32];

    if (objc < 2 || objc > 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "fileName ?zipFileName?");
	return TCL_ERROR;
    }
    if (objc > 2) {
	rsrc = objv[2];
    }

    chan = Wrap_Open(interp, rsrc, Tcl_GetString(objv[1]));
    if (!chan) {
	return TCL_ERROR;
    }

    Tcl_RegisterChannel(interp, chan);
    return TCL_OK;
}
/*
 *-------------------------------------------------------------------------
 *
 * wrapInfo --
 *
 *	This procedure is invoked to process the "::wrap::info" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 * *-------------------------------------------------------------------------
 */

static int wrapInfo(clientData, interp, objc, objv)
    ClientData clientData;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST objv[];
{
    int index;
    Tcl_Obj *rsrc = resource;
    static char *switches[] = {"files", (char *) NULL
    };
	        
    enum {
            INFO_FILES
    };

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
	return TCL_ERROR;
    }
    if (Tcl_GetIndexFromObj(interp, objv[1], switches, "option", 0, &index)
	    != TCL_OK) {
	return TCL_ERROR;
    }
    switch (index) {
	case INFO_FILES:
	    if (objc == 3) {
		rsrc = objv[2];
	    } else if (objc != 2) {
		Tcl_WrongNumArgs(interp, 2, objv, "?fileName?");
		return TCL_ERROR;
	    }
	    return Wrap_Files(interp, rsrc);
    }

    return TCL_OK;
}

static void wrapDeleteProc(ClientData clientData)
{
    Tcl_DecrRefCount(resource);
}
