/* 
 * wrapRsrc.c --
 *
 *	This file contains the code for the "wrapRsrcType", object type.
 *
 * Copyright (c) 1999      Jan Nijtmans
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "wrap.h"

#define WRAP_SHORT(buf, num) (buf[num] + (buf[num+1]<<8))
#define WRAP_LONG(buf, num) (WRAP_SHORT(buf,num) + (buf[num+2]<<16) + (buf[num+3]<<24))

/*
 * Prototypes for procedures defined later in this file:
 */

static void		FreeRsrcInternalRep _ANSI_ARGS_((Tcl_Obj *listPtr));
static int		SetRsrcFromAny _ANSI_ARGS_((Tcl_Interp *interp,
			    Tcl_Obj *objPtr));

/*
 * The structure below defines the rsrc Tcl object type by means of
 * procedures that can be invoked by generic object code.
 */

Tcl_ObjType wrapRsrcType = {
    "rsrc",				/* name */
    FreeRsrcInternalRep,		/* freeIntRepProc */
    (Tcl_DupInternalRepProc *) NULL,	/* dupIntRepProc */
    (Tcl_UpdateStringProc *) NULL,	/* updateStringProc */
    SetRsrcFromAny			/* setFromAnyProc */
};


/*
 *----------------------------------------------------------------------
 *
 * FreeRsrcInternalRep --
 *
 *	Deallocate the storage associated with a rsrc object's internal
 *	representation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees rsrcPtr's internal representation and sets rsrcPtr's
 *	internalRep.otherValuePtr to NULL. Decrements the ref counts
 *	of all element objects, which may free them.
 *
 *----------------------------------------------------------------------
 */

static void
FreeRsrcInternalRep(rsrcPtr)
    Tcl_Obj *rsrcPtr;		/* Rsrc object with internal rep to free. */
{
    Tcl_HashTable *table = (Tcl_HashTable *) rsrcPtr->internalRep.otherValuePtr;
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;

    entry = Tcl_FirstHashEntry(table, &search);
    while (entry) {
	/* ckfree((char *) Tcl_GetHashValue(entry)); */
    	entry = Tcl_NextHashEntry(&search);
    }

    Tcl_DeleteHashTable(table);
    ckfree((char *) table);
}

/*
 *----------------------------------------------------------------------
 *
 * SetRsrcFromAny --
 *
 *	Attempt to generate a list internal form for the Tcl object
 *	"objPtr".
 *
 * Results:
 *	The return value is TCL_OK or TCL_ERROR. If an error occurs during
 *	conversion, an error message is left in the interpreter's result
 *	unless "interp" is NULL.
 *
 * Side effects:
 *	If no error occurs, a list is stored as "objPtr"s internal
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static int
SetRsrcFromAny(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to convert. */
{
    Tcl_ObjType *oldTypePtr;
    int i, num, pos, start;
    unsigned char buffer[1024];
    char *string = Tcl_GetString(objPtr);
    Tcl_HashTable *table;
    Tcl_HashEntry *entry;
    Tcl_Channel chan;

    chan = Tcl_OpenFileChannel(interp, string, "r", 0);
    if (!chan) {
	return TCL_ERROR;
    }
    if (Tcl_SetChannelOption(interp, chan, "-translation", "binary") != TCL_OK) {
	return TCL_ERROR;
    }
    if (Tcl_SetChannelOption(interp, chan, "-encoding", "binary") != TCL_OK) {
	return TCL_ERROR;
    }

    pos = Tcl_Seek(chan, -22, SEEK_END);
    Tcl_Read(chan, buffer, 22);

    if (memcmp(buffer, "\120\113\05\06", 4)) {
    	Tcl_AppendResult(interp, "No resource found", NULL);
	return TCL_ERROR;
    }

    num = WRAP_SHORT(buffer,8);
    pos -= WRAP_LONG(buffer,12);
    start = pos - WRAP_LONG(buffer,16);

    Tcl_Seek(chan, pos, SEEK_SET);

    table = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(table, TCL_STRING_KEYS);

    while (num--) {
	int length1, length2, header;
	Tcl_Read(chan, buffer, 46);
	if (memcmp(buffer, "\120\113\01\02", 4)) {
	    Tcl_AppendResult(interp, "central header mismatch", NULL);
	    return TCL_ERROR;
	}

	length1 = WRAP_SHORT(buffer,28);
	length2 = WRAP_SHORT(buffer,30) + WRAP_SHORT(buffer,32);
	header = start + WRAP_LONG(buffer,42);

	Tcl_Read(chan, buffer, length1);	
	buffer[length1] = 0;
	entry = Tcl_CreateHashEntry(table, buffer, &i);
	/* value = (ClientData) ckalloc(size) */
	Tcl_SetHashValue(entry, (ClientData) header);

	while (length2>0) {
	    Tcl_Read(chan, buffer, 4);
	    length1 = WRAP_SHORT(buffer, 2);
	    Tcl_Read(chan, buffer+4, length1);
	    length2 -= length1 + 4;
	    if (!memcmp(buffer, "\106\113\023\0\115\104\65", 7)) {
		/* found MD5 signature */
		/* int i;
		for(i=0;i<16;i++) {
		    printf("%02X", buffer[i+7]);
		}
		printf("\n"); */
	    }
	}
    }
    Tcl_Close(interp, chan);

    /*
     * Free the old internalRep before setting the new one. We do this as
     * late as possible to allow the conversion code, in particular
     * Tcl_GetString, to use that old internalRep.
     */

    oldTypePtr = objPtr->typePtr;
    if ((oldTypePtr != NULL) &&	(oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = (VOID *) table;
    objPtr->typePtr = &wrapRsrcType;
    return TCL_OK;
}


int
Wrap_Files(interp, rsrc)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *rsrc;		/* The object to convert. */
{
    Tcl_Obj *result, *obj;
    Tcl_HashTable *table;
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;

    if (Tcl_ConvertToType(interp, rsrc, &wrapRsrcType) != TCL_OK) {
	return TCL_ERROR;
    }

    table = (Tcl_HashTable *) rsrc->internalRep.otherValuePtr;
    entry = Tcl_FirstHashEntry(table, &search);
    result = Tcl_NewObj();
    while (entry) {
	obj = Tcl_NewStringObj(Tcl_GetHashKey(table, entry), -1);
	Tcl_ListObjAppendElement(interp, result, obj);
    	entry = Tcl_NextHashEntry(&search);
    }
    Tcl_SetObjResult(interp, result);
    return TCL_OK;
}
